import threading
import logging

from pyatem.protocol import AtemProtocol


class HardwareThread(threading.Thread):
    def __init__(self, config):
        threading.Thread.__init__(self)
        self.name = 'hw.' + str(config['id'])
        self.config = config
        self.switcher = None
        self.stop = False
        self.status = 'init'

    def run(self):
        logging.info('HardwareThread run')
        self.status = 'connecting...'
        if self.config['address'] == 'usb':
            self.switcher = AtemProtocol(usb='auto')
        else:
            self.switcher = AtemProtocol(ip=self.config['address'])
        self.switcher.on('connected', self.on_connected)
        self.switcher.on('change', self.on_change)
        self.switcher.on('disconnected', self.on_disconnected)
        self.switcher.connect()
        while not self.stop:
            self.switcher.loop()

    def get_status(self):
        if self.status == 'connected':
            name = self.switcher.mixerstate["product-name"].name
            fw = self.switcher.mixerstate["firmware-version"].version
            self.status += f' ({name} fw {fw})'
        return self.status

    def on_connected(self):
        self.status = 'connected'
        logging.info('Initial state sync complete')

    def on_disconnected(self):
        self.status = 'lost connection'
        logging.error('Lost connection with the hardware')

    def on_change(self, key, value):
        pass



# Copyright 2022 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
from setuptools import setup

setup(
    name='openswitcher-proxy',
    version='0.13.0',
    packages=['openswitcher_proxy'],
    url='https://git.sr.ht/~martijnbraam/pyatem',
    license='LGPL3',
    author='Martijn Braam',
    author_email='martijn@brixit.nl',
    description='Proxy for ATEM switchers',
    requires=['pyatem', 'tomli >= 1.1.0 ; python_version < "3.11"'],
    extras_require={
        'midi': ['python-rtmidi'],
        'mqtt': ['paho-mqtt'],
    },
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: POSIX :: Linux',
    ],
    entry_points={
        'console_scripts': ['openswitcher-proxy=openswitcher_proxy.__main__:main'],
    },
)



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import math

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk


class LogAdjustment(Gtk.Adjustment):
    __gtype_name__ = 'LogAdjustment'

    def __init__(self, value, lower, upper, step_increment, page_increment, page_size, exp=False, range=60):
        super(Gtk.Adjustment, self).__init__()
        self.set_value(value)
        self.set_lower(lower)
        self.set_upper(upper)
        self.set_step_increment(step_increment)
        self.set_page_increment(page_increment)
        self.set_page_size(page_size)
        self.coeff = 10 ** (range / 20)
        self.exp = exp

        self.lower = lower
        self.upper = upper

    def to_normalized(self, value):
        upper = self.get_upper() - self.get_page_size()
        adj_range = upper - self.get_lower()
        return (value - self.get_lower()) / adj_range

    def from_normalized(self, value):
        upper = self.get_upper() - self.get_page_size()
        lower = self.get_lower()
        adj_range = upper - lower
        val = (value * adj_range) + lower
        return val

    def set_value_log(self, value):
        value = self.to_normalized(value)
        if self.exp:
            val = (math.exp((math.log(self.coeff + 1) * value)) - 1) / self.coeff
        else:
            val = math.log((value * self.coeff) + 1) / math.log(self.coeff + 1)
        self.set_value(self.from_normalized(val))

    def get_value_log(self):
        value = self.get_value()
        value = self.to_normalized(value)
        if self.exp:
            val = math.log((value * self.coeff) + 1) / math.log(self.coeff + 1)
        else:
            val = (math.exp((math.log(self.coeff + 1) * value)) - 1) / self.coeff
        return self.from_normalized(val)



import argparse
from pyatem.protocol import AtemProtocol
from pyatem.hexdump import hexdump

switcher = None
result = {}


def on_connected():
    global switcher
    global testqueue
    global prepqueue
    global stats_start

    print("Connection successful")
    model = switcher.mixerstate['product-name']
    print(f"Detected hardware: {model.name}")
    fw = switcher.mixerstate['firmware-version']
    print(f"Firmware: {fw}")
    print()
    print("Topology:")
    top = switcher.mixerstate['topology']
    print(top)
    print(hexdump(top.raw))
    exit(0)


def on_disconnected():
    print("Hardware has disconnected")


def run(device):
    global switcher
    print(f"Connecting to {device}...")
    switcher = AtemProtocol(device)
    switcher.on('connected', on_connected)
    switcher.on('disconnected', on_disconnected)
    switcher.connect()
    while True:
        switcher.loop()


def main():
    parser = argparse.ArgumentParser(description="Find non-implemented fields in the connected hardware")
    parser.add_argument('device', help="Device ip address or 'usb'")
    args = parser.parse_args()
    run(args.device)


if __name__ == '__main__':
    main()



import re
import threading
import logging
import json
from functools import partial
from json import JSONDecodeError

from .error import DependencyError
from .frontend_httpapi import FieldEncoder
import pyatem.command as commandmodule

try:
    import paho.mqtt.client as mqtt
    from paho.mqtt.subscribeoptions import SubscribeOptions
except ModuleNotFoundError:
    mqtt = None


class MqttFrontendThread(threading.Thread):
    def __init__(self, config, threadlist):
        threading.Thread.__init__(self)
        if mqtt is None:
            raise DependencyError("The paho-mqtt library is not available")
        self.name = 'mqtt.' + str(config['host'])
        self.config = config
        self.threadlist = threadlist
        self.hw_name = self.config['hardware'].split(',')
        self.topic = self.config['topic'] if 'topic' in self.config else "atem/{hardware}/{field}"
        self.client = None
        self.status = 'initializing...'
        self.error = None
        self.readonly = not self.config.get('allow-writes', False)
        self.subscribe = self.config['topic-subscribe'] if 'topic-subscribe' in self.config else self.topic

        regex = self.subscribe.replace('{hardware}', r'(?P<hardware>[^/]+)')
        regex = regex.replace('{field}', r'(?P<field>.+)')
        self.topic_re = re.compile(regex)

    def run(self):
        logging.info('MQTT frontend run')
        host, port = self.config['host'].split(':')
        port = int(port)
        self.client = mqtt.Client(client_id=f'atem-{self.name}', userdata=self, protocol=mqtt.MQTTv5)
        self.client.on_connect = lambda client, userdata, flags, rc, props: self.on_mqtt_connect(flags, rc, props)
        self.client.on_message = lambda client, userdata, msg: self.on_mqtt_message(msg)
        logging.info(f'connecting to {host}:{port}')
        try:
            self.client.connect(host, port, keepalive=3)
        except Exception as e:
            logging.error(f'Could not connect to the MQTT server at {host}:{port}')
            logging.error(e)
            self.error = f'could not connect to {host}:{port}'
            self.status = 'error'
            return
        for hw in self.hw_name:
            sw = self.threadlist['hardware'][hw].switcher

            # Hook into the events for the registered switchers and update the mqtt topic
            sw.on('connected', partial(self.on_switcher_connected, hw))
            sw.on('disconnected', partial(self.on_switcher_disconnected, hw))
            sw.on('change', partial(self.on_switcher_changed, hw))

            if self.threadlist['hardware'][hw].status == 'connected':
                # Hardware is already connected at this point, re-generate the initial data
                self.on_switcher_connected(hw)

        self.client.loop_forever()

    def on_switcher_changed(self, hw, field, value):
        raw = json.dumps(value, cls=FieldEncoder)
        topic = self.topic.format(hardware=hw, field=field)
        self.client.publish(topic, raw)

    def on_switcher_connected(self, hw):
        self.on_switcher_changed(hw, 'status', {'upstream': True})
        sw = self.threadlist['hardware'][hw].switcher
        items = list(sw.mixerstate.items())
        for field, value in items:
            self.on_switcher_changed(hw, field, value)

    def on_switcher_disconnected(self, hw):
        self.on_switcher_changed(hw, 'status', {'upstream': False})

    def on_mqtt_connect(self, flags, rc, properties):
        self.status = 'running'
        logging.info(f'MQTT: connected ({rc})')
        if not self.readonly:
            sub = self.subscribe.replace('{hardware}', '+').replace('{field}', '#')
            logging.info(f'Subscribing to topic {sub}')
            self.client.subscribe((sub, SubscribeOptions(noLocal=True)))

    def on_mqtt_message(self, msg):
        if self.readonly:
            logging.error('MQTT writes disabled')
            return

        match = self.topic_re.match(msg.topic)
        if not match:
            logging.error(f'MQTT: malformed command topic: {msg.topic}')

        hw = match.group('hardware')
        fieldname = match.group('field')

        if hw not in self.hw_name:
            logging.error(f'MQTT: not handling writes for "{hw}"')
            return

        if self.threadlist['hardware'][hw].status != "connected":
            logging.error(f'MQTT: writing to disconnected device "{hw}"')
            return

        classname = fieldname.title().replace('-', '') + "Command"
        if not hasattr(commandmodule, classname):
            logging.error(f'MQTT: unrecognized command {fieldname}')
            return
        try:
            arguments = json.loads(msg.payload)
        except JSONDecodeError as e:
            logging.error('received malformed payload, need a JSON dict')
            return
        if not isinstance(arguments, dict):
            logging.error(f'MQTT: mailformed payload, needs a JSON dict')
            return
        for key in arguments:
            try:
                arguments[key] = int(arguments[key])
            except:
                pass

        source_name_fault = False
        if 'source' in arguments:
            inputs = self.threadlist['hardware'][hw].switcher.inputs
            if not isinstance(arguments['source'], int) and arguments['source'] not in inputs:
                source_name_fault = True
            if arguments['source'] in inputs:
                arguments['source'] = inputs[arguments['source']]
        try:
            cmd = getattr(commandmodule, classname)(**arguments)
            self.threadlist['hardware'][hw].switcher.send_commands([cmd])
        except Exception as e:
            if source_name_fault:
                logging.error(f'MQTT: invalid source name: {arguments["source"]}')
            else:
                logging.error(f'MQTT: cannot write {fieldname}: {str(e)}')

    def get_status(self):
        if self.status == 'error':
            return f'{self.status}, {self.error}'
        if self.readonly:
            return self.status + ' (readonly)'
        else:
            return self.status + ' (writable)'



import base64
from http.server import BaseHTTPRequestHandler


class AuthRequestHandler(BaseHTTPRequestHandler):
    def verify_auth(self):
        if self.config['auth']:
            ok = False
            if 'Authorization' in self.headers:
                auth = base64.b64encode(f'{self.config["username"]}:{self.config["password"]}'.encode()).decode()
                correct = f'Basic {auth}'
                if self.headers['Authorization'] == correct:
                    ok = True

            if not ok:
                self.send_response(401)
                self.send_header('WWW-Authenticate', 'Basic realm="OpenSwitcher"')
                self.send_header('Content-type', 'text/html')
                self.end_headers()
                self.wfile.write('Authentication required'.encode())
                return False
        return True



import threading
import logging
import http.server
from functools import partial

from openswitcher_proxy.frontend import AuthRequestHandler


class StatusRequestHandler(AuthRequestHandler):
    def __init__(self, config, threadpool, *args, **kwargs):
        self.config = config
        self.threadpool = threadpool
        super().__init__(*args, **kwargs)

    def do_GET(self):
        if not self.verify_auth():
            return

        if self.path == '/favicon.ico':
            self.send_response(404)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            return

        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self.wfile.write("<!DOCTYPE html>\n<title>OpenSwitcher proxy status</title>".encode())
        self.wfile.write("<h1>Status</h1>".encode())
        self.wfile.write("<h2>Hardware</h2>".encode())
        if 'hardware' in self.threadpool:
            self.wfile.write(
                '<table border="1"><tr><th>id</th><th>label</th><th>address</th><th>status</th></tr>'.encode())
            for hwid in self.threadpool['hardware']:
                hardware = self.threadpool['hardware'][hwid]
                self.wfile.write('<tr>'.encode())
                self.wfile.write(f'<td>{hwid}</td>'.encode())
                self.wfile.write(f'<td>{hardware.config["label"]}</td>'.encode())
                self.wfile.write(f'<td>{hardware.config["address"]}</td>'.encode())
                self.wfile.write(f'<td>{hardware.get_status()}</td>'.encode())
                self.wfile.write('</tr>'.encode())
            self.wfile.write('</table>'.encode())
        else:
            self.wfile.write("No hardware is defined".encode())

        self.wfile.write("<h2>Frontends</h2>".encode())
        if 'frontend' in self.threadpool:
            self.wfile.write(
                '<table border="1"><tr><th>bind</th><th>type</th><th>auth</th><th>status</th></tr>'.encode())
            for bind in self.threadpool['frontend']:
                frontend = self.threadpool['frontend'][bind]
                self.wfile.write('<tr>'.encode())
                self.wfile.write(f'<td>{bind}</td>'.encode())
                self.wfile.write(f'<td>{frontend.config["type"]}</td>'.encode())
                self.wfile.write(f'<td>{frontend.config["auth"]}</td>'.encode())
                self.wfile.write(f'<td>{frontend.get_status()}</td>'.encode())
                self.wfile.write('</tr>'.encode())
            self.wfile.write('</table>'.encode())
        else:
            self.wfile.write("No hardware is defined".encode())


class StatusFrontendThread(threading.Thread):
    def __init__(self, config, threadlist):
        threading.Thread.__init__(self)
        self.name = 'status.' + str(config['bind'])
        self.config = config
        self.threadlist = threadlist
        self.stop = False
        self.server = None

    def run(self):
        logging.info('Status frontend run')
        host, port = self.config['bind'].split(':')
        address = (host, int(port))
        logging.info(f'binding to {address}')

        handler = partial(StatusRequestHandler, self.config, self.threadlist)

        self.server = http.server.HTTPServer(address, handler)
        with self.server:
            self.server.serve_forever()

    def handler(self, *args):
        print(args)

    def get_status(self):
        return 'running'



import argparse
import os
from datetime import datetime
import itertools

import yaml

from pyatem.protocol import AtemProtocol
import pyatem.field as fieldmodule
import pyatem.command as commandmodule

switcher = None
current = None
send_time = None
in_prep = True
prepqueue = []
testqueue = []
rejectbuffer = []

stats_executed = 0
stats_start = None


def on_change(key, contents):
    global switcher
    global stats_executed
    global rejectbuffer
    delta = datetime.now() - send_time

    if in_prep:
        send_next()
        return

    expect = current['expect']

    if delta.total_seconds() > 2:
        print("\nTimeout while waiting for field response")
        print("Sent:")
        print(f"  {current['send']['cmd']}:")
        for arg in current['send']:
            if arg == 'cmd':
                continue
            print(f"    {arg} = {current['send'][arg]}")
        print(f"Expected: {expect['field']}")
        print("Received:")
        for k in rejectbuffer:
            print(f"  {k}")
        exit(2)

    if key != expect['field']:
        rejectbuffer.append(key)
        return

    rejectbuffer = []

    for argname in expect:
        if argname == 'field':
            continue

        if getattr(contents, argname) != expect[argname]:
            print("Sent:")
            print(f"  {current['send']['cmd']}:")
            for arg in current['send']:
                if arg == 'cmd':
                    continue
                print(f"    {arg} = {current['send'][arg]}")
            print(f"\nMismatch on {argname}: expected {expect[argname]}, got {getattr(contents, argname)}")
            exit(2)

    stats_executed += 1
    print(".", end='')
    send_next()


def send_next():
    global switcher
    global prepqueue
    global in_prep
    global testqueue
    global current
    global send_time


    if len(testqueue) == 0:
        print("\n")
        print(f"Executed tests: {stats_executed}")
        print(f"Test duration: {datetime.now() - stats_start}")
        print("Tests completed successfully")
        exit(0)

    if len(prepqueue) > 0:
        test = prepqueue.pop(0)
    else:
        if in_prep:
            print("Running tests...")
            in_prep = False
        test = testqueue.pop(0)
    current = test

    classname = test['send']['cmd'].title().replace('-', '') + "Command"
    arguments = test['send'].copy()
    del arguments['cmd']

    if hasattr(commandmodule, classname):
        cmd = getattr(commandmodule, classname)(**arguments)
    else:
        print(f"Unknown command: {classname}")
        exit(2)

    send_time = datetime.now()
    switcher.send_commands([cmd])


def on_connected():
    global switcher
    global testqueue
    global prepqueue
    global stats_start

    print("Connection successful")
    model = switcher.mixerstate['product-name']
    print(f"Detected hardware: {model.name}")

    switcher.on('change', on_change)

    testfile = f'testset/{model.name}.yml'
    if not os.path.isfile(testfile):
        print("No test set defined for this hardware")
        exit(1)

    with open(testfile) as handle:
        tests = yaml.load(handle.read(), yaml.Loader)
    print(f"Loaded {len(tests['tests'])} test definitions")
    print("Setting initial state...")

    for prep in tests['prepare']:
        prepqueue.append({
            'send': prep,
        })

    for test in tests['tests']:
        if 'parameter' in test:
            sets = []
            for pname in test['parameter']:
                s = []
                for val in test['parameter'][pname]:
                    s.append((pname, val))
                sets.append(s)
            for combo in itertools.product(*sets):
                parameters = {}
                for arg in combo:
                    parameters[arg[0]] = arg[1]

                expectparam = parameters.copy()
                parameters.update(test['send'])
                expectparam.update(test['expect'])

                testqueue.append({
                    'send': parameters,
                    'expect': expectparam
                })
        else:
            testqueue.append(test)

    stats_start = datetime.now()
    send_next()


def on_disconnected():
    print("Hardware has disconnected")


def run_tests(device):
    global switcher
    print(f"Connecting to {device}...")
    switcher = AtemProtocol(device)
    switcher.on('connected', on_connected)
    switcher.on('disconnected', on_disconnected)
    switcher.connect()
    while True:
        switcher.loop()


def main():
    parser = argparse.ArgumentParser(description="pyatem unit test tool")
    parser.add_argument('device', help="Device ip address or 'usb'")
    args = parser.parse_args()
    run_tests(args.device)


if __name__ == '__main__':
    main()



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only

# Needs to be a global for decorators
_callbacks = {}


def call_fields(name, self, data):
    if name in _callbacks:
        for f in _callbacks[name]:
            f(self, data)


def field(name):
    def wrapper(func):
        global _callbacks
        if name not in _callbacks:
            _callbacks[name] = []
        _callbacks[name].append(func)
        return func

    return wrapper



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import math

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk


class AdjustmentEntry(Gtk.Entry):
    __gtype_name__ = 'AdjustmentEntry'

    def __init__(self, adjustment, display_min, display_max):
        super(Gtk.Entry, self).__init__()
        self.adjustment = None
        self.updating_model = False
        self.display_min = display_min
        self.display_max = display_max
        self.debounce = False
        self.set_adjustment(adjustment)
        self.get_style_context().add_class('adjustmententry')
        self.display = 0

    def set_adjustment(self, adjustment):
        self.adjustment = adjustment
        self.adjustment.connect('value-changed', self.adj_changed)
        self.adj_changed()

    def get_adjustment(self):
        return self.adjustment

    def _remap(self, value, old_min, old_max, new_min, new_max):
        return ((value - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min

    def _update(self, *args):
        self.updating_model = True
        text = "{:.2f}".format(self.display)
        # Entry.set_text() segfaults GTK3 occasionally
        self.set_text(text)
        self.updating_model = False
        self.debounce = False

    def adj_changed(self, *args):
        if hasattr(self.adjustment, 'get_value_log'):
            value = self.adjustment.get_value_log()
        else:
            value = self.adjustment.get_value()
        self.display = self._remap(value, self.adjustment.get_lower(), self.adjustment.get_upper(), self.display_min,
                                   self.display_max)

        if not self.debounce:
            self.debounce = True
            GLib.timeout_add(500, self._update)

    def do_activate(self):
        if not isinstance(self, AdjustmentEntry):
            return
        if self.updating_model:
            return

        try:
            value = float(self.get_text())
        except:
            return

        value = self._remap(value, self.display_min, self.display_max, self.adjustment.get_lower(),
                            self.adjustment.get_upper())

        if hasattr(self.adjustment, 'set_value_log'):
            self.adjustment.set_value_log(value)
        else:
            self.adjustment.set_value(value)



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import logging
import struct

from pyatem.transfer import TransferTask, TransferQueueFlushed
from pyatem.transport import UdpProtocol, Packet, UsbProtocol, TcpProtocol, ConnectionReady
from pyatem.command import LockCommand, TransferDownloadRequestCommand, TransferAckCommand, \
    TransferUploadRequestCommand, TransferDataCommand, TransferFileDataCommand, PartialLockCommand, TimeRequestCommand
from pyatem.media import rle_decode
import pyatem.field as fieldmodule


class AtemProtocol:
    STRUCT_FIELD = struct.Struct('!H2x 4s')

    FIELDNAME_PRETTY = {
        '_ver': 'firmware-version',
        '_pin': 'product-name',
        '_top': 'topology',
        'Time': 'time',
        'TCCc': 'time-config',
        'TcLk': 'timecode-lock',
        '_MeC': 'mixer-effect-config',
        '_MvC': 'multiviewer-config',
        '_FAC': 'fairlight-audio-config',
        '_VMC': 'video-mode-capability',
        'MvVM': 'multiview-video-mode-capability',
        '_MAC': 'macro-config',
        '_DVE': 'dve-capabilities',
        'Powr': 'power-status',
        '_mpl': 'mediaplayer-slots',
        'VidM': 'video-mode',
        'InPr': 'input-properties',
        'PrgI': 'program-bus-input',
        'PrvI': 'preview-bus-input',
        'TrSS': 'transition-settings',
        'TrPr': 'transition-preview',
        'TrPs': 'transition-position',
        'TMxP': 'transition-mix',
        'TDpP': 'transition-dip',
        'TWpP': 'transition-wipe',
        'TDvP': 'transition-dve',
        'TStP': 'transition-stinger',
        'KeOn': 'key-on-air',
        'KeBP': 'key-properties-base',
        'KeLm': 'key-properties-luma',
        'KACk': 'key-properties-advanced-chroma',
        'KACC': 'key-properties-advanced-chroma-colorpicker',
        'KePt': 'key-properties-pattern',
        'KeDV': 'key-properties-dve',
        'KeFS': 'key-properties-fly',
        'KKFP': 'key-properties-fly-keyframe',
        'DskB': 'dkey-properties-base',
        'DskP': 'dkey-properties',
        'DskS': 'dkey-state',
        'FtbP': 'fade-to-black',
        'FtbS': 'fade-to-black-state',
        'ColV': 'color-generator',
        'AuxS': 'aux-output-source',
        'MPfe': 'mediaplayer-file-info',
        'MPCE': 'mediaplayer-selected',
        'MPSp': 'mediaplayer-space',
        'MPCS': 'mediaplayer-clip-source',
        'MPAS': 'mediaplayer-audio-source',
        'RCPS': 'mediaplayer-clip-status',
        'AMMO': 'audio-mixer-master-properties',
        'AMmO': 'audio-mixer-monitor-properties',
        'AMTl': 'audio-mixer-tally',
        'FASP': 'fairlight-strip-properties',
        'FAMP': 'fairlight-master-properties',
        'FMPP': 'fairlight-properties',
        'FMHP': 'fairlight-headphones',
        'FAMS': 'fairlight-solo',
        'AIXP': 'fairlight-expander-properties',
        'AICP': 'fairlight-compressor-properties',
        'AILP': 'fairlight-limiter-properties',
        'MOCP': 'fairlight-master-compressor-properties',
        'AMLP': 'fairlight-master-limiter-properties',
        'ATMP': 'talkback-mixer-properties',
        'TMIP': 'talkback-mixer-input-properties',
        'MMOP': 'mix-minus-output-properties',
        '_TlC': 'tally-config',
        'TlIn': 'tally-index',
        'TlSr': 'tally-source',
        'FMTl': 'fairlight-tally',
        'MPrp': 'macro-properties',
        'MRPr': 'macro-play-status',
        'MRcS': 'macro-record-status',
        'CCst': 'camera-control-settings',
        'AiVM': 'auto-input-video-mode',
        'FASD': 'fairlight-strip-delete',
        'FAIP': 'fairlight-audio-input',
        'AMIP': 'audio-input',
        'RTMR': 'recording-duration',
        'RTMD': 'recording-disk',
        'RTMS': 'recording-status',
        'RMSu': 'recording-settings',
        'SRSU': 'streaming-service',
        'STAB': 'streaming-audio-bitrate',
        'StRS': 'streaming-status',
        'SRST': 'streaming-time',
        'SRSS': 'streaming-stats',
        'SAth': 'streaming-authentication',
        'AEBP': 'atem-eq-band-properties',
        'AMBP': 'atem-master-eq-band-properties',
        'MvPr': 'multiviewer-properties',
        'MvIn': 'multiviewer-input',
        'VuMC': 'multiviewer-vu',
        'VuMo': 'multiviewer-vu-opacity',
        'SaMw': 'multiviewer-safe-area',
        'StMv': 'multiviewer-safe-area-type',
        'LKOB': 'lock-obtained',
        'FTDa': 'file-transfer-data',
        'LKST': 'lock-state',
        'FTDE': 'file-transfer-error',
        'FTDC': 'file-transfer-data-complete',
        'FTCD': 'file-transfer-continue-data',
        'AMLv': 'audio-meter-levels',
        'FMLv': 'fairlight-meter-levels',
        'FDLv': 'fairlight-master-levels',
        'CCdP': 'camera-control-data-packet',
        '*XFC': 'transfer-complete',
        '_SSC': 'supersource-config',
        'SSrc': 'supersource-properties',
        'SSBP': 'supersource-box-properties',
        'V3sl': 'sdi-3g-level',
        'RXMS': 'hyperdeck-settings',
        'RXCP': 'hyperdeck-status',
        'RXSS': 'hyperdeck-storage',
        'RXCC': 'hyperdeck-clip-count',
        'DCPV': 'displayclock-properties',
        'DSTV': 'displayclock-set-time',
        'FEna': 'fade-to-black-enabled',
    }

    FIELDNAME_UNIQUE = {
        'mixer-effect-config': struct.Struct('>B'),
        'input-properties': struct.Struct('>H'),
        'program-bus-input': struct.Struct('>B'),
        'preview-bus-input': struct.Struct('>B'),
        'transition-preview': struct.Struct('>B'),
        'transition-position': struct.Struct('>B'),
        'transition-mix': struct.Struct('>B'),
        'transition-dip': struct.Struct('>B'),
        'transition-wipe': struct.Struct('>B'),
        'transition-dve': struct.Struct('>B'),
        'transition-stinger': struct.Struct('>B'),
        'fairlight-strip-properties': struct.Struct('>H'),
        'macro-properties': struct.Struct('>H'),
        'fairlight-audio-input': struct.Struct('>H'),
        'audio-input': struct.Struct('>H'),
        'key-on-air': struct.Struct('>BB'),
        'key-properties-base': struct.Struct('>BB'),
        'key-properties-luma': struct.Struct('>BB'),
        'key-properties-pattern': struct.Struct('>BB'),
        'key-properties-dve': struct.Struct('>BB'),
        'key-properties-fly': struct.Struct('>BB'),
        'key-properties-fly-keyframe': struct.Struct('>BBB'),
        'dkey-properties-base': struct.Struct('>B'),
        'dkey-properties': struct.Struct('>B'),
        'dkey-state': struct.Struct('>B'),
        'fade-to-black': struct.Struct('>B'),
        'fade-to-black-state': struct.Struct('>B'),
        'color-generator': struct.Struct('>B'),
        'aux-output-source': struct.Struct('>B'),
        'mediaplayer-file-info': struct.Struct('>xxH'),
        'mediaplayer-selected': struct.Struct('>B'),
        'atem-eq-band-properties': struct.Struct('>H14xB'),
        'multiviewer-properties': struct.Struct('>B'),
        'multiviewer-input': struct.Struct('>BB'),
        'multiviewer-vu': struct.Struct('>BB'),
        'multiviewer-safe-area': struct.Struct('>BB'),
        'camera-control-data-packet': struct.Struct('>BBB'),
        'supersource-box-properties': struct.Struct('>BB'),
        'supersource-properties': struct.Struct('>B'),
    }

    def __init__(self, ip=None, port=9910, usb=None):
        if ip is None and usb is None:
            raise ValueError("Need either an ip or usb port")
        if ip is not None:
            if ip.startswith('tcp://'):
                self.transport = TcpProtocol(url=ip)
            else:
                self.transport = UdpProtocol(ip, port)
        else:
            self.transport = UsbProtocol(usb)

        self.log = logging.getLogger('AtemProtocol')
        self.transport.queue_callback = self.queue_callback
        self.mixerstate = {}
        self.callbacks = {}
        self.inputs = {}
        self.callback_idx = 1
        self.connected = False

        self.locks = {}
        self.mode = None
        self.transfer_queue = {}
        self.transfer_id = 42
        self.transfer_buffer = b''
        self.transfer_buffer2 = []
        self.transfer = None
        self.transfer_requested = False
        self.transfer_packets = 0
        self.transfer_budget = []

    @classmethod
    def usb_exists(cls):
        return UsbProtocol.device_exists()

    def connect(self):
        self.log.debug('Starting connection')
        self.transport.connect()

    def loop(self):
        self.log.debug('Waiting for data packet...')
        packet = self.transport.receive_packet()
        if packet is None:
            # Disconnected from hardware
            if self.connected:
                self._raise('disconnected')
                self.mixerstate = {}
            self.connected = False
            return
        if isinstance(packet, ConnectionReady):
            self.connected = True
            self.send_commands([TimeRequestCommand()])
            self._raise('connected')
            return
        self.connected = True
        if isinstance(packet, TransferQueueFlushed):
            self._queue_flushed()
            return
        try:
            for fieldname, data in self.decode_packet(packet.data):
                self.save_field_data(fieldname, data)
        except ConnectionError:
            print("Encountered protocol corruption, closing connection")
            self._raise('disconnected')
            self.mixerstate = {}
            self.connected = False

    def on(self, event, callback):
        if event not in self.callbacks:
            self.callbacks[event] = {}
        self.callbacks[event][self.callback_idx] = callback
        self.callback_idx += 1
        return self.callback_idx - 1

    def off(self, event, callback_id):
        if event not in self.callbacks:
            return
        del self.callbacks[event][callback_id]

    def get_link_quality(self):
        return self.transport.get_link_quality()

    def _raise(self, event, *args, **kwargs):
        if event in self.callbacks:
            for cbidx in self.callbacks[event]:
                self.callbacks[event][cbidx](*args, **kwargs)

    def decode_packet(self, data):
        offset = 0
        while offset < len(data):
            datalen, cmd = self.STRUCT_FIELD.unpack_from(data, offset)

            # A zero length header is not possible, this occurs when the transport layer has corruption, mark the
            # connection closed to restart and recover state
            if datalen == 0:
                raise ConnectionError()

            raw = data[offset + 8:offset + datalen]
            yield (cmd, raw)
            offset += datalen

    def save_field_data(self, fieldname, contents):
        raw = contents
        key = fieldname.decode()
        if key in self.FIELDNAME_PRETTY:
            key = self.FIELDNAME_PRETTY[key]
            classname = key.title().replace('-', '') + "Field"
            if hasattr(fieldmodule, classname):
                contents = getattr(fieldmodule, classname)(contents)

        if key == 'CapA':
            return

        if key == 'lock-obtained':
            self.log.info('Got lock for {}'.format(contents.store))
            self.locks[contents.store] = True
            self._transfer_trigger(contents.store)
            return
        elif key == 'lock-state':
            if contents.state:
                # Ignore lock aquired messages from other clients
                return
            if contents.store in self.locks and self.locks[contents.store]:
                # Remove the lock if we held it
                del self.locks[contents.store]
            self.log.debug(contents)
            return
        elif key == 'file-transfer-continue-data':
            self.transfer_budget = contents
            old = self.transfer_budget.size
            self.transfer_budget.size = self.transfer_budget.size // 8 * 8
            if old != self.transfer_budget.size:
                self.log.debug(f"Adjusted transfer chunk size from {old} to {self.transfer_budget.size}")
            self._queue_chunks()
            return
        elif key == 'file-transfer-data':
            if contents.transfer == self.transfer.tid:
                self.transfer_packets += 1
                self.transfer_buffer += contents.data
                if self.transfer_packets % 20 == 0:
                    total_size = self.mixerstate['video-mode'].get_pixels() * 4
                    transfer_progress = len(self.transfer_buffer) / total_size
                    self._raise('transfer-progress', self.transfer.store, self.transfer.slot, transfer_progress)
                # The 0 should be the transfer slot, but it seems it's always 0 in practice
                self.send_commands([TransferAckCommand(self.transfer.tid, 0)])
            else:
                self.log.error('Got file transfer data for wrong transfer id')
            return
        elif key == 'file-transfer-error':
            self.log.error(f"file-transfer-error: {str(contents)}")
            self.transfer_requested = False
            if contents.status == 1:
                # Status is try-again
                self.log.debug('Retrying transfer')
                self._transfer_trigger(self.transfer.store, retry=True)
            elif contents.status == 5:
                self.locks[self.transfer.store] = False
                self._transfer_trigger(self.transfer.store, retry=True)
            return
        elif key == 'file-transfer-data-complete':
            self.log.debug('Transfer complete')
            if self.transfer is None:
                self.log.warning("Got FTDC without transfer active")
                return
            if contents.transfer != self.transfer.tid:
                return
            # Remove current item from the transfer queue
            queue = self.transfer_queue[self.transfer.store]
            self.transfer_queue[self.transfer.store] = queue[1:]

            if self.transfer.upload:
                self._raise('upload-done', self.transfer.store, self.transfer.slot)
                self.transfer_requested = False
            else:
                # Assemble the buffer
                data = self.transfer_buffer
                self.transfer_buffer = b''
                self.transfer_requested = False

                # Decompress the buffer if needed
                if self.transfer.store == 0:
                    data = rle_decode(data)

                self._raise('download-done', self.transfer.store, self.transfer.slot, data)

            # Start next transfer in the queue
            self._transfer_trigger(self.transfer.store)
            return
        elif key == 'transfer-complete':
            self.log.debug('Proxy transfer complete')

            # Remove current item from the transfer queue
            queue = self.transfer_queue[contents.store]
            self.transfer_queue[contents.store] = queue[1:]

            if contents.upload:
                self._raise('upload-done', contents.store, contents.slot)
            else:
                # TODO: Implement proxy download
                pass
            # Start next transfer in the queue
            self._transfer_trigger(self.transfer.store)
            return
        elif key == 'key-properties-dve':
            preset = contents.mini_preset_key
            if preset is not None:
                self._raise('pip-preset', preset)

        if key in self.FIELDNAME_UNIQUE:
            idxes = self.FIELDNAME_UNIQUE[key].unpack_from(raw, 0)
            if key not in self.mixerstate:
                self.mixerstate[key] = {}

            # Fairlight strips have weird numbering that's harder to parse here, read it back from the class
            if hasattr(contents, 'strip_id'):
                idxes = list(idxes)
                idxes[0] = contents.strip_id
                idxes = tuple(idxes)

            unique = self.make_unique_dict(contents, idxes)
            self.mixerstate[key] = self.recursive_merge(self.mixerstate[key], unique)
            self._raise('change:' + key + ':' + str(idxes[0]), contents)
            self._raise('change:' + key + ':*', contents)
        else:
            self.mixerstate[key] = contents
            self._raise('change:' + key, contents)
        if key == 'input-properties':
            self.inputs[contents.short_name] = contents.index

        if key == 'InCm':
            self.transport.mark_next_connected = True
            if isinstance(self.transport, TcpProtocol):
                self._raise('connected')
        self._raise('change', key, contents)

    def make_unique_dict(self, content, path):
        result = {}
        if len(path) == 1:
            result[path[0]] = content
        else:
            result[path[0]] = self.make_unique_dict(content, path[1:])
        return result

    def recursive_merge(self, d1, d2):
        '''update first dict with second recursively'''
        if not isinstance(d2, dict):
            return d2
        for k, v in d1.items():
            if k in d2:
                d2[k] = self.recursive_merge(v, d2[k])
        d1.update(d2)
        return d1

    def send_commands(self, commands):
        data = b''
        for command in commands:
            data += command.get_command()

        if len(data) > 1300:
            raise ValueError("Command list too long for UDP packet")

        self.send_raw(data)

    def send_raw(self, data):
        packet = Packet()
        packet.flags = UdpProtocol.FLAG_RELIABLE
        packet.data = data
        self.transport.send_packet(packet)

    def queue_callback(self, remaining, size):
        if not self.transfer:
            return

        self.transfer.send_done += size
        fraction = self.transfer.send_done / self.transfer.send_length
        self._raise('upload-progress', self.transfer.store, self.transfer.slot, fraction * 100, self.transfer.send_done,
                    self.transfer.send_length)

    def download(self, store, index):
        self.log.info("Queue download of {}:{}".format(store, index))
        if store not in self.transfer_queue:
            self.transfer_queue[store] = []
        self.transfer_queue[store].append(TransferTask(store, index))
        self._transfer_trigger(store)

    def upload(self, store, index, data, compress=True, compressed=False, name=None, description=None, size=None,
               task=None):
        self.log.info("Queue upload of {}:{}".format(store, index))
        if store not in self.transfer_queue:
            self.transfer_queue[store] = []

        if task is None:
            task = TransferTask(store, index, upload=True)
            task.data = data
            task.send_length = len(data)
            task.name = name
            task.description = description
            if compressed:
                uncompressed = rle_decode(data)
                task.data = uncompressed
                task.calculate_hash()
                task.data = data
            else:
                task.calculate_hash()
            if compress:
                task.compress()
            elif compressed:
                task.data_length = len(rle_decode(data))

        self.log.info(f'New upload task is {len(task.data)} bytes, {task.data_length} uncompressed')

        if isinstance(self.transport, TcpProtocol):
            self.transport.upload(task)
        else:
            self.transfer_queue[store].append(task)
            self._transfer_trigger(store)

    def _queue_chunks(self):
        # Can't transfer without a chunk size
        if self.transfer_budget is None:
            self.log.error('Cannot transfer without chunk size')
            return

        # Only queue chunks if an upload is planned
        if self.transfer is None:
            self.log.error('No transfer scheduled')
            return

        if not self.transfer.upload:
            self.log.error('Current transfer is a download')
            return

        chunk_size = self.transfer_budget.size
        self.log.debug(f'Queue {self.transfer_budget.count} chunks of {chunk_size}')
        for i in range(0, self.transfer_budget.count):
            if len(self.transfer.data) == 0:
                break

            chunk = self.transfer.data[0:chunk_size]
            used = chunk_size
            if chunk[-8:] == b'\xFE\xFE\xFE\xFE\xFE\xFE\xFE\xFE':
                chunk = self.transfer.data[0:chunk_size - 8]
                used -= 8
            elif chunk[-16:-8] == b'\xFE\xFE\xFE\xFE\xFE\xFE\xFE\xFE':
                chunk = self.transfer.data[0:chunk_size - 16]
                used -= 16
            self.transfer.data = self.transfer.data[used:]

            self.transfer_budget.count -= 1
            if self.transfer_budget.count == 0:
                self.log.debug('Transfer budget ran out')
                self.transfer_budget = None

            cmd = TransferDataCommand(self.transfer.tid, chunk)
            packet = Packet()
            packet.flags = UdpProtocol.FLAG_RELIABLE
            packet.data = cmd.get_command()
            self.transport.queue_packet(packet)
        self.transport.queue_trigger()

    def _queue_flushed(self):
        self.log.info('Queue flushed')
        if len(self.transfer.data):
            self._queue_chunks()
            return
        self.log.info('Sending file metadata')
        cmd = TransferFileDataCommand(self.transfer.tid, self.transfer.hash,
                                      name=self.transfer.name, description=self.transfer.description)
        self.send_commands([cmd])

    def _transfer_trigger(self, store, retry=False):
        next = None

        self.log.info(f'transfer trigger for store {store} (retry={retry})')

        # Try the preferred queue
        if store in self.transfer_queue:
            if len(self.transfer_queue[store]) > 0:
                next = self.transfer_queue[store][0]

        # Try any queue
        if next is None:
            for store in self.transfer_queue:
                if len(self.transfer_queue[store]) > 0:
                    next = self.transfer_queue[store][0]
                    break

        self.log.info(f'next transfer: {next}')

        # All transfers done, clean locks
        if next is None:
            for lock in self.locks:
                if self.locks[lock]:
                    self.log.info('Releasing lock {}'.format(lock))
                    cmd = LockCommand(lock, False)
                    self.send_commands([cmd])
            return

        # Request a lock if needed
        if next.store != 0xffff and (next.store not in self.locks or not self.locks[next.store]):
            self.log.info('Requesting lock for {}'.format(next.store))
            cmd = PartialLockCommand(next.store, next.slot)
            self.send_commands([cmd])
            return

        # A transfer request is already running, don't start a new one
        if self.transfer_requested:
            self.log.info('Request already submitted, do nothing')
            return

        # Assign a transfer id and start the transfer
        if not retry:
            self.transfer_id += 1
        self.transfer = next
        self.transfer.tid = self.transfer_id

        if self.transfer.upload:
            cmd = TransferUploadRequestCommand(self.transfer.tid, self.transfer.store, self.transfer.slot,
                                               self.transfer.data_length, 1)
            self.log.info('Requesting upload to {}:{}'.format(next.store, next.slot))
        else:
            cmd = TransferDownloadRequestCommand(self.transfer.tid, self.transfer.store, self.transfer.slot)
            self.log.info('Requesting download of {}:{}'.format(next.store, next.slot))
        self.transfer_requested = True
        self.send_commands([cmd])



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import gi

from gtk_switcher.template_i18n import TemplateLocale

gi.require_version('Gtk', '3.0')

from gi.repository import Gtk, GObject


@TemplateLocale(resource_path='/nl/brixit/switcher/ui/mixeffect-aux.glade')
class AuxMixEffectBlock(Gtk.Grid):
    __gtype_name__ = 'AuxMixEffectBlock'

    dsk_box = Gtk.Template.Child()

    aux_name = Gtk.Template.Child()
    bus = Gtk.Template.Child()
    focus_dummy = Gtk.Template.Child()

    def __init__(self, index, name):
        super(Gtk.Grid, self).__init__()
        self.init_template()
        self.index = index
        self.aux_name.set_label(name)

    def set_class(self, widget, classname, state):
        if state:
            widget.get_style_context().add_class(classname)
        else:
            widget.get_style_context().remove_class(classname)

    def set_inputs(self, buttons):
        # Clear the existing buttons
        for child in self.bus:
            child.destroy()

        for top, row in enumerate(buttons):
            for left, button in enumerate(row):
                if button is None:
                    spacer = Gtk.Box()
                    spacer.set_size_request(4, 4)
                    spacer.source_index = -1
                    pspacer = Gtk.Box()
                    pspacer.set_size_request(4, 4)
                    pspacer.source_index = -1

                    self.bus.attach(spacer, left, top, 1, 1)
                    continue
                active = button.short_name != ""

                label = Gtk.Label(label=button.short_name)

                btn = Gtk.Button()
                btn.add(label)
                btn.source_index = button.index
                btn.set_sensitive(active)
                btn.set_size_request(48, 48)
                btn.get_style_context().add_class('bmdbtn')
                btn.connect('clicked', self.do_program_input_change)
                self.bus.attach(btn, left, top, 1, 1)

        self.bus.show_all()

    def source_change(self, source):
        for btn in self.bus:
            if btn.source_index == source:
                btn.get_style_context().add_class('program')
            else:
                btn.get_style_context().remove_class('program')

    @GObject.Signal(name="source-changed", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def source_changed(self, *args):
        pass

    def do_program_input_change(self, widget):
        self.focus_dummy.grab_focus()
        self.emit("source-changed", self.index, widget.source_index)



import argparse
import struct

import requests
import logging


def get(ip, endpoint):
    base = f'http://{ip}/admin/api/v1'
    return requests.get(f'{base}/{endpoint}').json()


def put(ip, endpoint, payload):
    base = f'http://{ip}/admin/api/v1'
    return requests.put(f'{base}/{endpoint}', data=payload)


def get_device_capability(ip, capability):
    capa = get(ip, 'capabilities')
    if 'response' not in capa:
        logging.fatal("Did not get a valid response to the capabilities request")

    if capability not in capa['response']:
        return 0

    version = capa['response'][capability]['version']
    return version


def get_setup_basic(ip):
    return get(ip, 'setupBasic')


def get_firmware_info(ip):
    return get(ip, 'firmware/info')


def get_device_info(ip):
    setupcap = get_device_capability(ip, 'setupBasic')
    if setupcap != 1:
        logging.fatal("Device does not support the setupBasic capability")
        exit(1)

    basic = get_setup_basic(ip)['response']
    logging.info(f"Device:   {basic['productName']}")
    logging.info(f"Name:     {basic['deviceName']}")
    logging.info(f"Firmware: {basic['software']}")

    fwcap = get_device_capability(ip, 'firmware')
    if fwcap == 0:
        logging.fatal("Device does not support HTTP firmware upgrades")
        exit(1)

    if fwcap != 1:
        logging.fatal(f"Unknown firmware protocol version: {fwcap}")
        exit(1)

    fwinfo = get_firmware_info(ip)['response']
    logging.info(f"Hardware: 0x{fwinfo['hardware version']}")
    logging.info(f"Software: 0x{fwinfo['software version']}")
    logging.info(f"Filename: data-{fwinfo['product id'].lower()}.bin")
    logging.info("")

    return fwinfo['product id']


def load_firmware(file):
    with open(file, 'rb') as handle:
        fwheader = handle.read(80)
        fwdata = handle.read()

    part = struct.unpack_from('>32s HH 4x I H 2x I 4x 2x 2x 4x 4x HH', fwheader)
    checksum = part[0]
    length = part[3]
    headersize = part[4]
    fwlength = part[5]
    vid, pid = part[6], part[7]
    hpid = hex(pid).upper()[2:]
    piep = len(fwdata)
    if (length - headersize) != fwlength:
        logging.fatal("Understanding of header format incomplete, exit for safety")
        exit(1)

    if len(fwdata) != (part[3] - 48):
        logging.fatal("Firmware file length doesn't match header length")
        exit(1)

    return hpid, fwdata


def update_firmware(ip, firmware):
    fpid, fwdata = firmware
    response = put(ip, 'firmware/updateStart', b'').json()
    if 'success' not in response:
        logging.fatal("updateStart did not have a success key")
        exit(1)
    if not response['success']:
        logging.fatal("updateStart did not return a success result")
        exit(1)
    if 'response' not in response:
        logging.fatal("updateStart did not have a response key")
        exit(1)
    if 'upload_name' not in response['response']:
        logging.fatal("updateStart did not return an upload_name")
        exit(1)
    upload_name = response['response']['upload_name']
    logging.info(f"Upload filename is {upload_name}")

    logging.info("Uploading firmware to device...")
    response = put(ip, f'firmware/upload/{upload_name}', fwdata)
    if response.status_code != 200:
        logging.error(f'Update failed, http status {response.status_code}')
        exit(1)
    logging.info("Finalizing firmware update...")
    response = put(ip, 'firmware/updateFinalize', b'').json()
    if 'success' not in response:
        logging.fatal("updateFinalize did not have a success key")
        exit(1)
    if not response['success']:
        logging.fatal("updateFinalize did not return a success result")
        exit(1)
    logging.info("Firmware update completed")


def main():
    parser = argparse.ArgumentParser(description="ATEM firmware tool")
    parser.add_argument('ip', help='IP address of the ATEM device')
    parser.add_argument('file', nargs='?', help='Firmware filename')
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO, format="%(message)s")

    if args.ip == "inspect":
        firmware = load_firmware(args.file)
        print(f"Firmware is for device {firmware[0]}")
        exit(0)

    product = get_device_info(args.ip)

    if args.file:
        firmware = load_firmware(args.file)
        logging.info("Everything seems to be ready to go, press enter to continue or ctrl+c to abort")
        input()
        if product != firmware[0]:
            logging.error(f"Product ID of device [{product}] does not match firmware file [{firmware[0]}]")
        update_firmware(args.ip, firmware)


if __name__ == '__main__':
    main()



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import queue
import socket
import os


class SocketQueue(queue.Queue):
    """
    This is a queue.Queue that's also a socket so it works with the select() call
    to await both a queue item and a network packet.
    """

    def __init__(self):
        super().__init__()

        if os.name == 'posix':
            self._putsocket, self._getsocket = socket.socketpair()
        else:
            # Compatibility on non-POSIX systems
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.bind(('127.0.0.1', 0))
            server.listen(1)
            self._putsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._putsocket.connect(server.getsockname())
            self._getsocket, _ = server.accept()
            server.close()

    def fileno(self):
        return self._getsocket.fileno()

    def put(self, item, **kwargs):
        super().put(item, **kwargs)
        self._putsocket.send(b'x')

    def get(self, **kwargs):
        self._getsocket.recv(1)
        return super().get(**kwargs)



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
from unittest import TestCase

from pyatem.hexdump import hexdump
from pyatem.media import rle_decode, rle_encode


class Test(TestCase):
    def _rle_loop_check(self, label, testdata):
        compressed = rle_encode(testdata)
        decoded = rle_decode(compressed)
        if decoded != testdata:
            print("Original:")
            hexdump(testdata)
            print("Compressed:")
            hexdump(compressed)
            print("Result:")
            hexdump(decoded)

        self.assertEqual(testdata, decoded, label)

    def test_rle_uncompressable(self):
        testdata = b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x02\x02\x02\x02\x02\x02\x02\x02'
        testdata += b'\x03\x03\x03\x03\x03\x03\x03\x03'
        testdata += b'\x04\x04\x04\x04\x04\x04\x04\x04'
        self._rle_loop_check('uncompressable', testdata)

    def test_rle_single_block(self):
        testdata = b''
        for i in range(1, 10):
            testdata += b'\x01\x01\x01\x01\x01\x01\x01\x01'
            self._rle_loop_check(f'single block {i}', testdata)

    def test_rle_double_block(self):
        testdata1 = b''
        testdata2 = b''
        for i in range(1, 10):
            testdata1 += b'\x01\x01\x01\x01\x01\x01\x01\x01'
            testdata2 += b'\x02\x02\x02\x02\x02\x02\x02\x02'
            self._rle_loop_check(f'double block {i}', testdata1 + testdata2)

    def test_rle_block_purge(self):
        testdata = b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x02\x02\x02\x02\x02\x02\x02\x02'
        self._rle_loop_check('end block purge', testdata)

    def test_rle_block_purge_double(self):
        testdata = b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x02\x02\x02\x02\x02\x02\x02\x02'
        testdata += b'\x02\x02\x02\x02\x02\x02\x02\x02'
        self._rle_loop_check('double end block purge', testdata)

    def test_rle_two_compressable_gap(self):
        testdata = b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\x01\x01\x01\x01\x01\x01\x01\x01'
        testdata += b'\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa'
        testdata += b'\x02\x02\x02\x02\x02\x02\x02\x02'
        testdata += b'\x02\x02\x02\x02\x02\x02\x02\x02'
        testdata += b'\x02\x02\x02\x02\x02\x02\x02\x02'
        testdata += b'\x02\x02\x02\x02\x02\x02\x02\x02'
        self._rle_loop_check('third block', testdata)



# Copyright 2022 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import ipaddress
import math
import os.path
import time

import usb.core
import usb.util
import struct

from pyatem.converters.lut import lut_to_bmd17, load_cube, lut_to_bmd33


class Field:
    def __init__(self, code, key, dtype, section, label, mapping=None, sys=False, ro=False):
        self.code = code
        self.key = key
        self.section = section
        self.dtype = dtype
        self.label = label
        self.sys = sys
        self.mapping = mapping
        self.ro = ro

        self.value = None
        self.widget = None

    def __repr__(self):
        return f'<Field {self.key} ({self.label})>'


class Option:
    def __init__(self, code, label):
        self.code = code
        self.label = label


class Converter:
    VENDOR = 0x1edb
    PRODUCT = 0
    NAME = "Unknown"
    FIELDS = []
    PROTOCOL = "label"

    STATUS_OK = 1
    STATUS_NEED_POWER = 2

    def __init__(self):
        self.handle = None
        self.serial = None

    @classmethod
    def is_plugged_in(cls, serial=None):
        if serial is not None:
            result = usb.core.find(idVendor=cls.VENDOR, idProduct=cls.PRODUCT, iSerialNumber=serial)
        else:
            result = usb.core.find(idVendor=cls.VENDOR, idProduct=cls.PRODUCT)
        return result is not None

    def enumerate(cls):
        result = []
        for dev in usb.core.find(idVendor=cls.VENDOR, idProduct=cls.PRODUCT, find_all=True):
            if dev.serial_number:
                result.append(dev.serial_number)
            else:
                result.append(f'{dev.bus}-{dev.address}')
        return result

    def connect(self, serial=None):
        if serial is not None:
            if '-' in serial:
                bus, address = serial.split('-')
                bus = int(bus)
                address = int(address)
                self.handle = usb.core.find(idVendor=self.VENDOR, idProduct=self.PRODUCT, bus=bus, address=address)
            else:
                self.handle = usb.core.find(idVendor=self.VENDOR, idProduct=self.PRODUCT, serial_number=serial)

        else:
            self.handle = usb.core.find(idVendor=self.VENDOR, idProduct=self.PRODUCT)

        self.handle.set_configuration(1)

    def get_name(self):
        # Fallback for devices that might not support renaming
        return self.NAME

    def get_status(self):
        return self.STATUS_OK

    def get_version(self):
        return None

    def get_value_raw(self, field):
        raise NotImplementedError()

    def set_value_raw(self, field, value):
        raise NotImplementedError()

    def get_value(self, field):
        raise NotImplementedError()

    def set_value(self, field, value):
        raise NotImplementedError()

    def get_state(self):
        result = {}
        for field in self.FIELDS:
            ret = self.get_value(field)
            field.value = ret
            result[field.key] = field
        return result

    def factory_reset(self):
        raise NotImplementedError()


class LabelProtoConverter(Converter):
    PROTOCOL = 'label'
    NAME_FIELD = "DeviceName"
    VERSION_FIELD = "ReleaseVersion"

    def get_name(self):
        return self._communicate(self.NAME_FIELD, sys=True).decode()

    def get_version(self):
        return self._communicate(self.VERSION_FIELD, sys=True).decode()

    def factory_reset(self):
        self.handle.ctrl_transfer(bmRequestType=0x40,
                                  bRequest=20,
                                  wValue=0,
                                  wIndex=0,
                                  data_or_wLength=0)

    def get_value_raw(self, field):
        if field.dtype == open:
            return None
        return self._communicate(field.key, field.sys)

    def get_value(self, field):
        raw = self.get_value_raw(field)
        if field.dtype == str:
            value = raw.split(b'\0')[0].decode()
        elif field.dtype == int:
            value = int.from_bytes(raw, byteorder='little')
        elif field.dtype == bool:
            value = int.from_bytes(raw, byteorder='little') > 0
        elif field.dtype == ipaddress.IPv4Address:
            value = ipaddress.IPv4Address(raw)
        elif field.dtype == open:
            value = None
        else:
            raise ValueError("Unknown type")

        if field.mapping == 'dB':
            if value > 0:
                value = 20 * math.log10(value / 64)
            else:
                value = float('-inf')
        return value

    def set_value_raw(self, field, value):
        self._communicate(field.key, field.sys, write=value)

    def set_value(self, field, value):
        if isinstance(value, bytes):
            self.set_value_raw(field, value)

        if field.dtype == int:
            value = value.to_bytes(length=(8 + (value + (value < 0)).bit_length()), byteorder='little')
        elif field.dtype == str:
            value = value.encode()
        elif field.dtype == bool:
            value = 255 if value else 0
            value = value.to_bytes(field.key[1], byteorder='little')

        self.set_value_raw(field, value)

    def _communicate(self, name, sys=False, write=None):
        ep_read = 0xa1 if sys else 0xc0
        ep_write = 0x21 if sys else 0x40
        req_ticket = 1 if sys else 10
        req_name = 2 if sys else 11
        req_read = 3 if sys else 12
        req_write = 4 if sys else 13

        ticket = self.handle.ctrl_transfer(bmRequestType=ep_read,
                                           bRequest=req_ticket,
                                           wValue=0,
                                           wIndex=0,
                                           data_or_wLength=2)

        ticket, = struct.unpack('<H', bytes(ticket))

        self.handle.ctrl_transfer(bmRequestType=ep_write,
                                  bRequest=req_name,
                                  wValue=ticket,
                                  wIndex=0,
                                  data_or_wLength=name.encode())

        if write is not None:
            large = len(write) > 4096
            while len(write) > 0:
                chunk = write[0:4096]
                write = write[4096:]
                self.handle.ctrl_transfer(bmRequestType=ep_write,
                                          bRequest=req_write,
                                          wValue=ticket,
                                          wIndex=0,
                                          data_or_wLength=chunk,
                                          timeout=10000 if large else None)
        else:
            return bytes(self.handle.ctrl_transfer(bmRequestType=ep_read,
                                                   bRequest=req_read,
                                                   wValue=ticket,
                                                   wIndex=0,
                                                   data_or_wLength=255))

    def set_lut(self, key, path):
        lut = load_cube(path)
        title = os.path.basename(path)
        title = '.'.join(title.split('.')[0:-1])
        stream = lut_to_bmd17(lut)

        self._communicate("LutData", False, stream)
        self._communicate("LutName", False, title)


class WValueProtoConverter(Converter):
    PROTOCOL = 'wValue'
    NAME_FIELD = 0x00C0
    VERSION_FIELD = 0x00B0
    LUTFIELD = False
    HAS_NAME = True
    NEEDS_POWER = False
    LUT_SIZE = 17

    REG_STATUS = 48
    CMD_CLEAR = 55
    REG_CLEARSTATUS = 56
    CMD_WRITE = 57

    """
    bRequest:
      82: setting write
      83: setting read
      160: is powered?
    """

    def get_name(self):
        if not self.HAS_NAME:
            name = self.NAME.replace('Blackmagic design ', '')
            name = name.replace('Mini', '').replace('Micro', '').replace('Converter', '')
            return name.strip()

        raw = bytes(self.handle.ctrl_transfer(bmRequestType=0xc1,
                                              bRequest=83,
                                              wValue=self.NAME_FIELD,
                                              wIndex=0,
                                              data_or_wLength=64))

        return raw.split(b'\0')[0].decode()

    def get_status(self):
        if not self.NEEDS_POWER:
            return Converter.STATUS_OK
        raw = bytes(self.handle.ctrl_transfer(bmRequestType=0xc1,
                                              bRequest=160,
                                              wValue=0x00,
                                              wIndex=0,
                                              data_or_wLength=1))
        if raw == b'\0':
            return Converter.STATUS_NEED_POWER
        return Converter.STATUS_OK

    def get_version(self):
        raw = bytes(self.handle.ctrl_transfer(bmRequestType=0xc1,
                                              bRequest=83,
                                              wValue=self.VERSION_FIELD,
                                              wIndex=0,
                                              data_or_wLength=7))

        return raw.split(b'\0')[0].decode()

    def get_value_raw(self, field):
        if field.dtype == open:
            return
        return bytes(self.handle.ctrl_transfer(bmRequestType=0xc1,
                                               bRequest=83,
                                               wValue=field.key[0],
                                               wIndex=0,
                                               data_or_wLength=field.key[1]))

    def get_value(self, field):
        raw = self.get_value_raw(field)
        if field.dtype == str:
            value = raw.split(b'\0')[0].decode()
        elif field.dtype == int:
            value = int.from_bytes(raw, byteorder='little')
        elif field.dtype == bool:
            value = int.from_bytes(raw, byteorder='little') > 0
        elif field.dtype == ipaddress.IPv4Address:
            value = ipaddress.IPv4Address(raw)
        elif field.dtype == open:
            value = None
        else:
            raise ValueError("Unknown type")

        if field.mapping == 'dB':
            if value > 0:
                value = 20 * math.log10(value / 64)
            else:
                value = float('-inf')
        return value

    def set_value_raw(self, field, value):
        self.handle.ctrl_transfer(bmRequestType=0x41,
                                  bRequest=82,
                                  wValue=field.key[0],
                                  data_or_wLength=value)

    def set_value(self, field, value):
        if isinstance(value, bytes):
            self.set_value_raw(field, value)

        if field.mapping == "dB":
            print(f"set {value} dB")
            value = float(value)
            value = int(round(math.pow(10, value / 20) * 64))

        if field.dtype == int:
            value = value.to_bytes(field.key[1], byteorder='little')
        elif field.dtype == str:
            # TODO(Peter): Fix me! This has some "fun" edge cases if unicode is
            # allowed...
            if len(value.encode()) > field.key[1] - 1:
                value = value[0:field.key[1] - 2].encode()
            else:
                value = value.encode()
            value += b'\0'
        elif field.dtype == bool:
            value = 255 if value else 0
            value = value.to_bytes(field.key[1], byteorder='little')
        elif field.dtype == ipaddress.IPv4Address:
            addr = ipaddress.IPv4Address(value)
            value = addr.packed

        self.set_value_raw(field, value)

    def _read(self, bRequest, length):
        return bytes(self.handle.ctrl_transfer(bmRequestType=0xc1,
                                               bRequest=bRequest,
                                               wValue=0,
                                               wIndex=0,
                                               data_or_wLength=length))

    def _write(self, bRequest, data):
        self.handle.ctrl_transfer(bmRequestType=0x41,
                                  bRequest=bRequest,
                                  wValue=0,
                                  wIndex=0,
                                  data_or_wLength=data)

    def _wait_on_status(self, status0=None, status1=None, status2=None, status3=None, status4=None, status5=None):
        # Wait for the LUT engine to be ready
        wanted = [status0, status1, status2, status3, status4, status5]
        for i in range(0, 20):
            status = self._read(self.REG_STATUS, 6)
            data = struct.unpack('>6B', status)

            fail = False
            for idx, s in enumerate(wanted):
                if s is not None and data[idx] != s:
                    fail = True
            if not fail:
                break

            time.sleep(0.5)
        else:
            raise TimeoutError("Status did not update")

    def _clear_region(self, address, length):
        self._write(self.CMD_CLEAR, struct.pack('>II', address, length))
        while True:
            status = self._read(self.REG_CLEARSTATUS, 15)
            part = struct.unpack('>II I BBB', status)
            # 0 address
            # 1 length
            # 2 clear ptr
            # 3 ?
            # 4 ?
            # 5 ?

            progress = ((part[2] - part[0]) / part[1]) * 100
            if part[3] == 0 and part[4] == 0 and part[5] == 0:
                break

            time.sleep(0.3)

    def _bulk_write(self, address, data):
        # position in blocks maybe?
        block = address >> 8
        self._write(self.CMD_WRITE, struct.pack('>I', block))

        self.handle.write(1, data)
        self.handle.write(1, b'')

    def _set_lut_33(self, key, path):
        address, length = key
        lut = load_cube(path)
        title = os.path.basename(path)
        title = '.'.join(title.split('.')[0:-1])
        stream = lut_to_bmd33(lut, title)

        self._wait_on_status(status4=0xFF, status5=0xFF)
        self._write(49, b'')
        self._write(52, b'')

        self._clear_region(address, length)
        self._bulk_write(address, stream)
        self._wait_on_status(status1=0x00, status2=0x00)
        self._write(50, b'')

    def _set_lut_17(self, key, path):
        address, length = key
        lut = load_cube(path)
        stream = lut_to_bmd17(lut)

        self._wait_on_status(status4=0xFF, status5=0xFF)
        self._write(49, b'')
        self._write(52, b'')
        self._clear_region(address, length)
        self._bulk_write(address, stream)
        self._wait_on_status(status1=0x00, status2=0x00)
        self._write(50, b'')

        # Write the new LUT name and finalize the upload
        self.set_value(Field(None, (0x0410, 64), str, '', 'LUT name'), struct.pack('>64s', lut.title.encode()))
        self.set_value(Field(None, (0x0400, 1), int, '', 'Unknown'), struct.pack('>B', 3))

    def set_lut(self, key, path):
        if self.LUT_SIZE == 17:
            self._set_lut_17(key, path)
        elif self.LUT_SIZE == 33:
            self._set_lut_33(key, path)
        else:
            raise ValueError("Unsupported LUT size")


class WIndexProtoConverter(Converter):
    PROTOCOL = 'wIndex'
    #    NAME_FIELD = 0x00C0
    #    VERSION_FIELD = 0x00B0
    #    LUTFIELD = False
    HAS_NAME = True
    NEEDS_POWER = False
    #    LUT_SIZE = 17

    REG_STATUS = 48
    CMD_CLEAR = 55
    REG_CLEARSTATUS = 56
    CMD_WRITE = 57

    """
    bRequest:
      215: setting write
      214: setting read
      225: is powered?
    """

    def get_name(self):
        if not self.HAS_NAME:
            name = self.NAME.replace('Blackmagic design ', '')
            name = name.replace('Mini', '').replace('Micro', '').replace('Converter', '')
            return name.strip()

        raise NotImplementedError()

    #        raw = bytes(self.handle.ctrl_transfer(bmRequestType=0xc1,
    #                                              bRequest=83,
    #                                              wValue=self.NAME_FIELD,
    #                                              wIndex=0,
    #                                              data_or_wLength=64))
    #
    #        return raw.split(b'\0')[0].decode()

    def get_status(self):
        if not self.NEEDS_POWER:
            return Converter.STATUS_OK
        raw = bytes(self.handle.ctrl_transfer(bmRequestType=0xc0,
                                              bRequest=225,
                                              wValue=0x0000,
                                              wIndex=0,
                                              data_or_wLength=1))
        if raw == b'\0':
            return Converter.STATUS_NEED_POWER
        return Converter.STATUS_OK

    def get_version(self):
        raise NotImplementedError()
        print("TODO: fixme!")

    #        raw = bytes(self.handle.ctrl_transfer(bmRequestType=0xc1,
    #                                              bRequest=83,
    #                                              wValue=self.VERSION_FIELD,
    #                                              wIndex=0,
    #                                              data_or_wLength=7))
    #
    #        return raw.split(b'\0')[0].decode()

    def get_value_raw(self, field):
        if field.dtype == open:
            return
        return bytes(self.handle.ctrl_transfer(bmRequestType=0xc0,
                                               bRequest=214,
                                               wValue=0,
                                               wIndex=field.key[0],
                                               data_or_wLength=field.key[1]))

    def get_value(self, field):
        raw = self.get_value_raw(field)
        if field.dtype == str:
            value = raw.split(b'\0')[0].decode()
        elif field.dtype == int:
            value = int.from_bytes(raw, byteorder='little')
        elif field.dtype == bool:
            value = int.from_bytes(raw, byteorder='little') > 0
        elif field.dtype == open:
            value = None
        else:
            raise ValueError("Unknown type")

        if field.mapping == 'dB':
            if value > 0:
                value = 20 * math.log10(value / 64)
            else:
                value = float('-inf')
        return value

    def set_value_raw(self, field, value):
        # Set seems to be URB Control wIndex of field ID ORed with the desired field
        # value with bmRequestType of 0x40
        self.handle.ctrl_transfer(bmRequestType=0x40,
                                  bRequest=215,
                                  wValue=0,
                                  # Or the field ID with the first byte of the
                                  # value in case there is more than one byte
                                  wIndex=field.key[0] | value[0],
                                  data_or_wLength=0),

    def set_value(self, field, value):
        if isinstance(value, bytes):
            self.set_value_raw(field, value)

        if field.mapping == "dB":
            print(f"set {value} dB")
            value = float(value)
            value = int(round(math.pow(10, value / 20) * 64))

        if field.dtype == int:
            value = value.to_bytes(field.key[1], byteorder='little')
        elif field.dtype == str:
            # TODO(Peter): Test me, this is probably more correct than what's
            # currently used, but untested and given it's a set...
            # TODO(Peter): Fix me! This has some "fun" edge cases if unicode is
            # allowed...
            # if len(value.encode()) > field.key[1] - 1:
            #     value = value[0:field.key[1] - 2].encode()
            # else:
            #     value = value.encode()
            # value += b'\0'
            value = value.encode()
        elif field.dtype == bool:
            value = 255 if value else 0
            value = value.to_bytes(field.key[1], byteorder='little')
        # TODO(Peter): Test me, this is probably correct, but untested and
        # given it's a set...
        # elif field.dtype == ipaddress.IPv4Address:
        #     addr = ipaddress.IPv4Address(value)
        #     value = addr.packed

        self.set_value_raw(field, value)

    def _read(self, bRequest, length):
        raise NotImplementedError()

    #        return bytes(self.handle.ctrl_transfer(bmRequestType=0xc1,
    #                                               bRequest=bRequest,
    #                                               wValue=0,
    #                                               wIndex=0,
    #                                               data_or_wLength=length))

    def _write(self, bRequest, data):
        raise NotImplementedError()

    #        self.handle.ctrl_transfer(bmRequestType=0x41,
    #                                  bRequest=bRequest,
    #                                  wValue=0,
    #                                  wIndex=0,
    #                                  data_or_wLength=data)

    def _wait_on_status(self, status0=None, status1=None, status2=None, status3=None, status4=None, status5=None):
        # Wait for the LUT engine to be ready
        wanted = [status0, status1, status2, status3, status4, status5]
        for i in range(0, 20):
            status = self._read(self.REG_STATUS, 6)
            data = struct.unpack('>6B', status)

            fail = False
            for idx, s in enumerate(wanted):
                if s is not None and data[idx] != s:
                    fail = True
            if not fail:
                break

            time.sleep(0.5)
        else:
            raise TimeoutError("Status did not update")

    def _clear_region(self, address, length):
        self._write(self.CMD_CLEAR, struct.pack('>II', address, length))
        while True:
            status = self._read(self.REG_CLEARSTATUS, 15)
            part = struct.unpack('>II I BBB', status)
            # 0 address
            # 1 length
            # 2 clear ptr
            # 3 ?
            # 4 ?
            # 5 ?

            progress = ((part[2] - part[0]) / part[1]) * 100
            if part[3] == 0 and part[4] == 0 and part[5] == 0:
                break

            time.sleep(0.3)

    def _bulk_write(self, address, data):
        # position in blocks maybe?
        block = address >> 8
        self._write(self.CMD_WRITE, struct.pack('>I', block))

        self.handle.write(1, data)
        self.handle.write(1, b'')

    def _set_lut_33(self, key, path):
        address, length = key
        lut = load_cube(path)
        title = os.path.basename(path)
        title = '.'.join(title.split('.')[0:-1])
        stream = lut_to_bmd33(lut, title)

        self._wait_on_status(status4=0xFF, status5=0xFF)
        self._write(49, b'')
        self._write(52, b'')

        self._clear_region(address, length)
        self._bulk_write(address, stream)
        self._wait_on_status(status1=0x00, status2=0x00)
        self._write(50, b'')

    def _set_lut_17(self, key, path):
        address, length = key
        lut = load_cube(path)
        stream = lut_to_bmd17(lut)

        self._wait_on_status(status4=0xFF, status5=0xFF)
        self._write(49, b'')
        self._write(52, b'')
        self._clear_region(address, length)
        self._bulk_write(address, stream)
        self._wait_on_status(status1=0x00, status2=0x00)
        self._write(50, b'')

        # Write the new LUT name and finalize the upload
        self.set_value(Field(None, (0x0410, 64), str, '', 'LUT name'), struct.pack('>64s', lut.title.encode()))
        self.set_value(Field(None, (0x0400, 1), int, '', 'Unknown'), struct.pack('>B', 3))

    def set_lut(self, key, path):
        if self.LUT_SIZE == 17:
            self._set_lut_17(key, path)
        elif self.LUT_SIZE == 33:
            self._set_lut_33(key, path)
        else:
            raise ValueError("Unsupported LUT size")


class AtemLegacyProtocol(Converter):
    PROTOCOL = 'AtemLegacy'
    NAME_FIELD = 0x0048

    def get_name(self):
        return self.get_value(Field((self.NAME_FIELD, 32), str, "Device", "Name")).decode()

    def get_version(self):
        return 'Unsupported'

    def get_value(self, field):
        if field.dtype == open:
            return

        result = b''
        for i in range(0, field.key[1]):
            char = bytes(self.handle.ctrl_transfer(bmRequestType=0xc0,
                                                   bRequest=214,
                                                   wValue=0,
                                                   wIndex=field.key[0] + i,
                                                   data_or_wLength=1))
            if char == '\0':
                break
            result += char
        return result

    def set_value(self, field, value):
        if field.dtype == str:
            if len(value) > field.key[1] - 1:
                value = value[0:field.key[1] - 2]
            value += b'\0'
        for i in range(0, len(value)):
            self.handle.ctrl_transfer(bmRequestType=0x40,
                                      bRequest=215,
                                      wIndex=field.key[0] + i,
                                      wValue=value[i])



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import collections

import gi

from pyatem.field import InputPropertiesField
from pyatem.hexdump import hexdump

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk, Pango

gi.require_version('Handy', '1')
from gi.repository import Handy


class DebuggerWindow:
    def __init__(self, parent, connection, application):
        self.application = application
        self.connection = connection

        builder = Gtk.Builder()
        builder.set_translation_domain("openswitcher")
        builder.add_from_resource('/nl/brixit/switcher/ui/debugger.glade')
        builder.connect_signals(self)
        css = Gio.resources_lookup_data("/nl/brixit/switcher/ui/style.css", 0)

        self.provider = Gtk.CssProvider()
        self.provider.load_from_data(css.get_data())

        self.window = builder.get_object("window")

        self.statemodel = builder.get_object("statemodel")
        self.state_tree = builder.get_object("state_tree")
        self.state_box = builder.get_object("state_box")
        self.events = builder.get_object("events")

        self.load_initial_state()

        self.apply_css(self.window, self.provider)
        self.connection.mixer.on('change', self.on_change)
        self.window.show_all()

    def apply_css(self, widget, provider):
        Gtk.StyleContext.add_provider(widget.get_style_context(),
                                      provider,
                                      Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        if isinstance(widget, Gtk.Container):
            widget.forall(self.apply_css, provider)

    def flatten(self, d, parent_key='', sep='_'):
        items = []
        for k, v in d.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, dict):
                items.extend(self.flatten(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    def load_initial_state(self):
        known = self.statemodel.append(None, ("Known",))
        unknown = self.statemodel.append(None, ("Unknown",))

        col_name = Gtk.TreeViewColumn("Field")
        render_name = Gtk.CellRendererText()
        col_name.pack_start(render_name, True)
        col_name.add_attribute(render_name, 'text', 0)

        self.state_tree.append_column(col_name)

        for fieldname in self.connection.mixer.mixerstate:
            field = self.connection.mixer.mixerstate[fieldname]
            if isinstance(field, bytes):
                parent = unknown
            else:
                parent = known
            self.statemodel.append(parent, (fieldname,))

    def on_state_selection_changed(self, widget, *args):
        model, model_iter = widget.get_selected()
        fieldname = model.get(model_iter, 0)[0]
        if fieldname == "Known" or fieldname == "Unknown":
            return

        for widget in self.state_box:
            self.state_box.remove(widget)

        field = self.connection.mixer.mixerstate[fieldname]
        if not isinstance(field, dict):
            field = {"0": field}

        flat = self.flatten(field)

        for key in flat:
            row = flat[key]
            self.state_box.add(self.make_field_widget(fieldname, row))
        self.apply_css(self.state_box, self.provider)
        self.state_box.show_all()

    def make_field_widget(self, name, field):
        frame = Gtk.Frame()
        frame.get_style_context().add_class('view')
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        frame.add(box)

        heading = Gtk.Label(name)
        heading.get_style_context().add_class("heading")
        heading.set_margin_start(10)
        heading.set_margin_end(10)
        heading.set_margin_top(10)
        heading.set_margin_bottom(10)

        box.add(heading)

        if isinstance(field, bytes):
            label = Gtk.Label(hexdump(field, result='return'))
            label.get_style_context().add_class('hexdump')
            label.set_margin_start(10)
            label.set_margin_end(10)
            label.set_margin_top(10)
            label.set_margin_bottom(10)
            box.add(label)
        else:
            grid = Gtk.Grid(column_spacing=10, row_spacing=3)
            grid.set_margin_bottom(10)
            i = 0
            for key in field.__dict__:
                if key == "raw":
                    continue
                keylabel = Gtk.Label(key, xalign=1.0)
                keylabel.set_hexpand(True)
                valuelabel = Gtk.Label(str(getattr(field, key)), xalign=0.0)
                valuelabel.set_line_wrap(True)
                valuelabel.set_hexpand(True)
                keylabel.get_style_context().add_class('dim-label')
                grid.attach(keylabel, 0, i, 1, 1)
                grid.attach(valuelabel, 1, i, 1, 1)
                i += 1

            box.add(grid)

        return frame

    def on_change(self, fieldname, field):
        if fieldname == 'time':
            return

        GLib.idle_add(self.on_change_uithread, fieldname, field)

    def on_change_uithread(self, fieldname, field):
        widget = self.make_field_widget(fieldname, field)
        widget.get_style_context().add_class('pulsein')
        self.apply_css(widget, self.provider)
        self.events.pack_end(widget, False, False, 0)
        self.events.show_all()



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
from gi.repository import Gtk, GObject, Gdk

from pyatem.command import KeyPropertiesDveCommand, DkeyMaskCommand, KeyPropertiesAdvancedChromaColorpickerCommand, \
    SupersourceBoxPropertiesCommand


class Region:
    def __init__(self, rid, layout):
        self.rid = rid
        self.layout = layout

        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0

        self.has_mask = False
        self.mask_top = 0
        self.mask_bottom = 0
        self.mask_left = 0
        self.mask_right = 0

        self.tally = False
        self.visible = True

    def _queue_draw(self):
        self.layout.da.queue_draw()

    def set_tally(self, onair):
        if onair == self.tally:
            return
        self.tally = onair
        self._queue_draw()

    def set_visible(self, visible):
        if visible == self.visible:
            return
        self.visible = visible
        self._queue_draw()

    def set_region(self, x, y, w, h):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self._queue_draw()

    def set_mask(self, top, bottom, left, right):
        self.has_mask = True
        self.mask_top = top / 18000
        self.mask_bottom = bottom / 18000
        self.mask_left = left / 32000
        self.mask_right = right / 32000
        self._queue_draw()

    def get_label(self):
        return "Error"

    def __repr__(self):
        return '<Region {} {},{} {}x{}>'.format(self.get_label(), self.x, self.y, self.w, self.h)

    def get_region_cairo(self, width, height):
        w = self.w / 16
        h = self.h / 9
        x = ((self.x + 16) / 32.0) - (w / 2)
        y = ((self.y + 9) / 18.0) - (h / 2)

        return x * width, y * height, w * width, h * height

    def get_mask_cairo(self, width, height):
        return self.mask_top, self.mask_bottom, self.mask_left, self.mask_right

    def _update_region(self, index, pos_x=None, pos_y=None, size_x=None, size_y=None):
        return []

    def _update_mask(self, index, left=None, right=None, top=None, bottom=None):
        return []


class UpstreamKeyRegion(Region):
    def get_label(self):
        return _("Upstream key {}").format(self.rid[1] + 1)

    def _update_region(self, index, pos_x=None, pos_y=None, size_x=None, size_y=None):
        if pos_x is not None:
            x = int((pos_x * 32 - 16) * 1000)
        if pos_y is not None:
            y = int((pos_y * 18 - 9) * 1000)

        w = None
        h = None
        if size_x is not None and size_y is not None:
            w = int(size_x * 1000)
            h = int(size_y * 1000)
        cmd = KeyPropertiesDveCommand(index=index, keyer=self.rid[1], pos_x=x, pos_y=y,
                                      size_x=w, size_y=h)
        return [cmd]

    def _update_mask(self, index, left=None, right=None, top=None, bottom=None):
        ml, mr, mt, mb = None, None, None, None
        if left is not None:
            ml = int(left * 32000)
        if right is not None:
            mr = int(right * 32000)
        if top is not None:
            mt = int(top * 18000)
        if bottom is not None:
            mb = int(bottom * 18000)

        cmd = KeyPropertiesDveCommand(index=index, keyer=self.rid[1], mask_top=mt, mask_bottom=mb, mask_left=ml,
                                      mask_right=mr)
        return [cmd]


class DownstreamKeyRegion(Region):
    def get_label(self):
        return _("Downstream keyer {}").format(self.rid[1] + 1)

    def _update_mask(self, index, left=None, right=None, top=None, bottom=None):
        ml, mr, mt, mb = None, None, None, None
        if left is not None:
            ml = int(left * 32000) - 16000
        if right is not None:
            mr = 16000 - int(right * 32000)
        if top is not None:
            mt = 9000 - int(top * 18000)
        if bottom is not None:
            mb = int(bottom * 18000) - 9000

        cmd = DkeyMaskCommand(index=self.rid[1], top=mt, bottom=mb, left=ml, right=mr)
        return [cmd]


class ColorPickerRegion(Region):
    def get_label(self):
        return _("Chroma picker {}").format(self.rid[1] + 1)

    def _update_region(self, index, pos_x=None, pos_y=None, size_x=None, size_y=None):
        x = None
        y = None
        if pos_x is not None:
            x = int((pos_x * 32 - 16) * 1000)
        if pos_y is not None:
            y = int((pos_y * 18 - 9) * 1000)

        size = None
        if size_x is not None:
            size = int(size_x * 16000)

        cmd = KeyPropertiesAdvancedChromaColorpickerCommand(index=index, keyer=self.rid[1], x=x, y=y, size=size)
        return [cmd]


class SuperSourceBoxRegion(Region):
    def get_label(self):
        return _("SuperSource Box {}").format(self.rid[1] + 1)

    def _update_region(self, index, pos_x=None, pos_y=None, size_x=None, size_y=None):
        if pos_x is not None:
            x = int((pos_x * 2 - 1) * 1600)
        if pos_y is not None:
            y = int((pos_y * 2 - 1) * 900)

        size = None
        if size_x is not None and size_y is not None:
            size = int(size_x * 1000)
        cmd = SupersourceBoxPropertiesCommand(index=0, box=self.rid[1], x=x, y=y,
                                              size=size)
        return [cmd]

    def _update_mask(self, index, left=None, right=None, top=None, bottom=None):
        ml, mr, mt, mb = None, None, None, None
        if left is not None:
            ml = int(left * 32000)
        if right is not None:
            mr = int(right * 32000)
        if top is not None:
            mt = int(top * 18000)
        if bottom is not None:
            mb = int(bottom * 18000)

        cmd = SupersourceBoxPropertiesCommand(index=0, box=self.rid[1], top=mt, bottom=mb, left=ml, right=mr)
        return [cmd]


class LayoutView(Gtk.Frame):
    __gtype_name__ = 'LayoutView'
    LAYER_USK = 1
    LAYER_DSK = 2
    LAYER_ACK = 3
    LAYER_SSB = 4

    def __init__(self, index, connection):
        super(Gtk.Frame, self).__init__()
        self.index = index
        self.connection = connection
        self.set_size_request(640, 480)
        self.set_margin_top(12)
        self.set_margin_bottom(12)
        self.set_margin_start(12)
        self.set_margin_end(12)
        self.get_style_context().add_class("view")

        self.da = Gtk.DrawingArea()
        self.add(self.da)

        self.da.set_events(
            self.da.get_events()
            | Gdk.EventMask.LEAVE_NOTIFY_MASK
            | Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.POINTER_MOTION_HINT_MASK
        )

        self.area_width = 0
        self.area_height = 0
        self.area_top = 0
        self.area_left = 0

        self.da.connect("draw", self.on_draw)
        self.da.connect("button-press-event", self.on_mouse_down)
        self.da.connect("button-release-event", self.on_mouse_up)
        self.da.connect("motion-notify-event", self.on_mouse_move)

        self.show_all()

        self.selected = None
        self.handle = None
        self.offset_x = None
        self.offset_y = None
        self.offset_alt_x = None
        self.offset_alt_y = None
        self.move_alt = False
        self.alt_x = None
        self.alt_y = None

        self.mask_top = 0
        self.mask_bottom = 0
        self.mask_left = 0
        self.mask_right = 0

        self.regions = {}

    def get(self, type, index):
        """
        Get a region by type and index, type is one of the LayoutView.LAYER_ constants.
        The region will be created internally if it did not exist yet.

        :rtype: Region
        """
        rid = (type, index)
        if rid not in self.regions:
            if type == LayoutView.LAYER_USK:
                region = UpstreamKeyRegion(rid, self)
            elif type == LayoutView.LAYER_DSK:
                region = DownstreamKeyRegion(rid, self)
            elif type == LayoutView.LAYER_ACK:
                region = ColorPickerRegion(rid, self)
            elif type == LayoutView.LAYER_SSB:
                region = SuperSourceBoxRegion(rid, self)
            self.regions[rid] = region
        return self.regions[rid]

    def on_context(self):
        menu = Gtk.Menu()
        for rid in self.regions:
            region = self.regions[rid]
            mi = Gtk.CheckMenuItem(region.get_label())
            mi.set_active(region.visible)
            menu.append(mi)
            mi.show()
            mi.rid = rid
            mi.connect('toggled', self.on_context_toggled)
        menu.popup(None, None, None, None, 0, Gtk.get_current_event_time())

    def on_context_toggled(self, widget, *args):
        self.regions[widget.rid].visible = widget.get_active()
        self.da.queue_draw()

    def on_mouse_down(self, widget, event):

        if event.button == 3:
            self.on_context()
            return

        if event.button != 1:
            return

        # grr cairo coordinates
        event.y = self.area_height - event.y - self.area_top
        event.x = event.x - self.area_left
        self.move_alt = event.state & Gdk.ModifierType.CONTROL_MASK > 0

        if self.selected is None:
            hits = []
            for rid in self.regions:
                region = self.regions[rid]
                if not region.visible:
                    continue
                x, y, w, h = region.get_region_cairo(self.area_width, self.area_height)

                if x < event.x < (x + w):
                    if y < event.y < (y + h):
                        hits.append((rid, w * h))

            if len(hits) == 0:
                return

            # Sort all hits on area of the region, always pick the smallest one so it's possible to select overlapping
            # regions naturally
            sorted_hits = list(sorted(hits, key=lambda k: k[1]))
            self.selected = sorted_hits[0][0]
            self.da.queue_draw()
            return

        region = self.regions[self.selected]
        x, y, w, h = region.get_region_cairo(self.area_width, self.area_height)
        mtop, mbottom, mleft, mright = region.get_mask_cairo(self.area_width, self.area_height)

        if (x - 5) < event.x < (x + 5):
            if (y - 5) < event.y < (y + 5):
                self.handle = 'bl'
                self.offset_x = x + w
                self.offset_y = y + h
                return

        if (x - 5) < event.x < (x + 5):
            if (y - 5 + h) < event.y < (y + 5 + h):
                self.handle = 'tl'
                self.offset_x = x + w
                self.offset_y = y
                return

        if (x - 5 + w) < event.x < (x + 5 + w):
            if (y - 5) < event.y < (y + 5):
                self.handle = 'br'
                self.offset_x = x
                self.offset_y = y + h
                return

        if (x - 5 + w) < event.x < (x + 5 + w):
            if (y - 5 + h) < event.y < (y + 5 + h):
                self.handle = 'tr'
                self.offset_x = x
                self.offset_y = y
                return

        if (x - 5 + (mleft * w)) < event.x < (x + 5 + (mleft * w)):
            if (y - 15 + (h / 2)) < event.y < (y + 15 + (h / 2)):
                self.handle = 'ml'
                self.offset_x = x
                return

        if (x - 5 + w - (mright * w)) < event.x < (x + 5 + w - (mright * w)):
            if (y - 15 + (h / 2)) < event.y < (y + 15 + (h / 2)):
                self.handle = 'mr'
                self.offset_x = x + w
                return

        if (x - 15 + (w / 2)) < event.x < (x + 15 + (w / 2)):
            if (y - 5 + h - (mtop * h)) < event.y < (y + 5 + h - (mtop * h)):
                self.handle = 'mt'
                self.offset_y = y + h
                return

        if (x - 15 + (w / 2)) < event.x < (x + 15 + (w / 2)):
            if (y - 5 + (mbottom * h)) < event.y < (y + 5 + (mbottom * h)):
                self.handle = 'mb'
                self.offset_y = y
                return

        if x < event.x < (x + w):
            if y < event.y < (y + h):
                self.handle = 'pos'
                if self.move_alt:
                    self.offset_alt_x = event.x
                    self.offset_alt_y = event.y
                    self.alt_x = (x + (w / 2))
                    self.alt_y = (y + (h / 2))
                self.offset_x = (x + (w / 2)) - event.x
                self.offset_y = (y + (h / 2)) - event.y
                return

        self.selected = None
        self.da.queue_draw()

    def on_mouse_up(self, widget, *args):
        self.handle = None
        self.da.queue_draw()

    def on_mouse_move(self, widget, event):
        if self.handle is None or self.selected is None:
            return

        # grr cairo coordinates
        event.y = self.area_height - event.y - self.area_top
        event.x = event.x - self.area_left

        region = self.regions[self.selected]
        x, y, w, h = region.get_region_cairo(self.area_width, self.area_height)

        if self.handle == 'pos':
            if self.move_alt:
                # Alternate move moves the source inside the mask area instead of the key itself

                # Vertical axis first
                new_left = min(1.0, max(0, ((event.x - self.offset_alt_x) / w) + self.mask_left))
                new_right = min(1.0, max(0, self.mask_right + (self.mask_left - new_left)))

                # Horizontal axis
                new_bottom = min(1.0, max(0, ((event.y - self.offset_alt_y) / h) + self.mask_bottom))
                new_top = min(1.0, max(0, self.mask_top + (self.mask_bottom - new_bottom)))

                # Move the flying key in the opposite direction to compensate
                new_x = self.alt_x + (self.offset_alt_x - event.x)
                new_y = self.alt_y + (self.offset_alt_y - event.y)

                mask_cmd = self.on_mask_update(self.selected, left=new_left, right=new_right, top=new_top,
                                               bottom=new_bottom, exec=False)
                pos_cmd = self.on_region_update(self.selected, pos_x=new_x, pos_y=new_y, exec=False)
                self.connection.mixer.send_commands([pos_cmd, mask_cmd])
            else:
                new_x = event.x + self.offset_x
                new_y = event.y + self.offset_y
                self.on_region_update(self.selected, pos_x=new_x, pos_y=new_y)
        elif self.handle.startswith("m"):
            if self.handle == "ml":
                new_left = min(1.0, max(0, (event.x - self.offset_x) / w))
                self.on_mask_update(self.selected, left=new_left)
            elif self.handle == "mr":
                new_right = min(1.0, max(0, (self.offset_x - event.x) / w))
                self.on_mask_update(self.selected, right=new_right)
            elif self.handle == "mt":
                new_top = min(1.0, max(0, (self.offset_y - event.y) / h))
                self.on_mask_update(self.selected, top=new_top)
            elif self.handle == "mb":
                new_bottom = min(1.0, max(0, (event.y - self.offset_y) / h))
                self.on_mask_update(self.selected, bottom=new_bottom)
        else:
            new_x = (self.offset_x + event.x) / 2
            new_y = (self.offset_y + event.y) / 2
            new_w = max(self.offset_x, event.x) - min(self.offset_x, event.x)
            new_h = max(self.offset_y, event.y) - min(self.offset_y, event.y)
            self.on_region_update(self.selected, pos_x=new_x, pos_y=new_y, size_x=new_w, size_y=new_h)

    def on_region_update(self, rid, pos_x=None, pos_y=None, size_x=None, size_y=None, exec=True):
        # Normalize the cairo coordinates again
        if pos_x is not None:
            pos_x = pos_x / self.area_width
        if pos_y is not None:
            pos_y = pos_y / self.area_height
        if size_x is not None:
            size_x = size_x / self.area_width
        if size_y is not None:
            size_y = size_y / self.area_height

        region = self.regions[rid]
        cmds = region._update_region(index=self.index, pos_x=pos_x, pos_y=pos_y, size_x=size_x, size_y=size_y)

        if exec:
            self.connection.mixer.send_commands(cmds)
        return cmds

    def on_mask_update(self, rid, left=None, right=None, top=None, bottom=None, exec=True):
        region = self.regions[rid]
        cmds = region._update_mask(index=self.index, left=left, right=right, top=top, bottom=bottom)
        if exec:
            self.connection.mixer.send_commands(cmds)
        return cmds

    def on_draw(self, widget, cr):

        # Calculate a 16:9 area inside the widget size
        width, height = widget.get_allocated_width(), widget.get_allocated_height()
        top = 0
        left = 0
        if width / height > 16 / 9:
            new_width = height * 16 // 9
            left = (width - new_width) // 2
            width = new_width
        else:
            new_height = width * 9 // 16
            top = (height - new_height)
            height = new_height

        self.area_width = width
        self.area_height = height
        self.area_top = top
        self.area_left = left

        # Clear
        cr.set_source_rgba(0.0, 0.0, 0.0, 0.0)
        cr.paint()

        # Draw frame
        cr.set_source_rgb(0.1, 0.1, 0.1)
        cr.rectangle(left, top, width, height)
        cr.fill()

        for rid in self.regions:
            region = self.regions[rid]
            if not region.visible:
                continue
            x, y, w, h = region.get_region_cairo(self.area_width, self.area_height)

            mtop, mbottom, mleft, mright = 0, 0, 0, 0
            if region.has_mask:
                mtop, mbottom, mleft, mright = region.get_mask_cairo(self.area_width, self.area_height)

            if region.tally:
                cr.set_source_rgb(1.0, 0, 0)
            else:
                cr.set_source_rgb(1.0, 1.0, 1.0)

            if mleft != 0:
                cr.move_to(x + left + (w * mleft), height - (y + top + h))
                cr.line_to(x + left + (w * mleft), height - (y + top))
            if mtop != 0:
                cr.move_to(x + left, height - (y + top + h) + (mtop * h))
                cr.line_to(x + left + w, height - (y + top + h) + (mtop * h))
            if mbottom != 0:
                cr.move_to(x + left, height - (y + top + h) + h - (mbottom * h))
                cr.line_to(x + left + w, height - (y + top + h) + h - (mbottom * h))
            if mright != 0:
                cr.move_to(x + left + w - (w * mright), height - (y + top + h))
                cr.line_to(x + left + w - (w * mright), height - (y + top))

            if self.selected == rid:
                cr.set_dash([2.0, 1.0])

            cr.rectangle(x + left, height - (y + top + h), w, h)
            cr.stroke()

            if self.selected == rid:
                # Position/Size handles on the corners
                cr.rectangle(x + left - 5, height - (y + top + h) - 5, 10, 10)
                cr.fill()
                cr.rectangle(x + left - 5 + w, height - (y + top + h) - 5, 10, 10)
                cr.fill()
                cr.rectangle(x + left - 5, height - (y + top) - 5, 10, 10)
                cr.fill()
                cr.rectangle(x + left - 5 + w, height - (y + top) - 5, 10, 10)
                cr.fill()
                cr.rectangle(x + left - 5 + (w / 2), height - (y + top + (h / 2)) - 5, 10, 10)
                cr.fill()

                # Mask handles on the sides
                if region.has_mask:
                    cr.rectangle(x + left - 4 + (w * mleft), height - (y + top + (h / 2)) - 15, 5, 30)
                    cr.fill()
                    cr.rectangle(x + left + w - (w * mright), height - (y + top + (h / 2)) - 15, 5, 30)
                    cr.fill()
                    cr.rectangle(x + left + (w / 2) - 15, height - (y + top + h - (mtop * h) + 4), 30, 5)
                    cr.fill()
                    cr.rectangle(x + left + (w / 2) - 15, height - (y + top + (mbottom * h)), 30, 5)
                    cr.fill()

            cr.move_to(x + left + 10, height - (y + top + h) + 20)
            cr.show_text(region.get_label())



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import argparse
import locale
import gettext
import logging
import os
import sys

import gi

from gtk_switcher.atemwindow import AtemWindow

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gio

gi.require_version('Handy', '1')
from gi.repository import Handy


class AtemLogFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    white = "\x1b[97;20m"
    blue = "\x1b[36;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = '[%(levelname)-8s %(threadName)-12s] %(name)-14s - %(message)s'

    FORMATS = {
        logging.DEBUG: blue + format + reset,
        logging.INFO: white + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class AtemApplication(Gtk.Application):
    def __init__(self, application_id, flags, args):
        Gtk.Application.__init__(self, application_id=application_id, flags=flags)
        self.connect("activate", self.new_window)
        self.args = args

        settings = Gtk.Settings.get_default()
        settings.set_property("gtk-application-prefer-dark-theme", True)

    def new_window(self, *args):
        AtemWindow(self, self.args)


def main(version):
    Handy.init()

    parser = argparse.ArgumentParser(description="ATEM Control panel")
    parser.add_argument('ip', help='ip-address of the switcher to connect to', nargs='?')
    parser.add_argument('--persist', action='store_true', help='save the new ip address')
    parser.add_argument('--verbose', action='store_true', help='Show more log messages')
    parser.add_argument('--debug', action='store_true', help='Display a lot of debugging info')
    parser.add_argument('--dump', help='dump data for specific packets', nargs='*')
    parser.add_argument('--view', choices=['switcher', 'media', 'audio', 'camera'], default='switcher',
                        help='default view to open when launching')
    args = parser.parse_args()

    ch = logging.StreamHandler()
    ch.setFormatter(AtemLogFormatter())

    if args.verbose:
        logging.basicConfig(level=logging.INFO, handlers=[ch])
    elif args.debug:
        logging.basicConfig(level=logging.DEBUG, handlers=[ch])
    else:
        logging.basicConfig(handlers=[ch])

    log = logging.getLogger('Launcher')
    log.info(f'OpenSwitcher version {version}')
    if os.path.isfile('atem.gresource'):
        log.info('using resources from cwd')
        resource = Gio.resource_load("atem.gresource")
        Gio.Resource._register(resource)

    app = AtemApplication("nl.brixit.Switcher", Gio.ApplicationFlags.FLAGS_NONE, args=args)
    app.run()


if __name__ == '__main__':
    locale.bindtextdomain("openswitcher", os.getenv('LOCALEDIR', '.'))
    gettext.install("openswitcher", os.getenv('LOCALEDIR', '.'))
    locale.textdomain("openswitcher")

    main('')
